"""Houses Option"""

from typing import Dict, Any
from dataclasses import dataclass
from json import dumps

@dataclass
class Option:
    """Option Type"""

    name: str
    value: Any

    @property
    def as_dict(self) -> Dict[str, Any]:
        """Gets the object as it's dict representation"""

        return {self.name: self.value}

    def to_json(self) -> str:
        """Gets the ojbect as it's JSON representation"""

        return dumps(self.as_dict)



"""Houses Header Option"""
from typing import Dict, Any

from pyrestsdk.type.model._option import Option


class HeaderOption(Option):
    """Header Option Type"""

    @property
    def as_dict(self) -> Dict[str, str]:
        return {str(self.name): str(self.value)}


from typing import Union

from base64 import b64encode

from pyrestsdk.credential._abstract_basic_credential import AbstractBasicCredential

class BasicCredential(AbstractBasicCredential):
    """Basic Authentication Credential Type"""

    def to_native_string(self, string: Union[str, bytes], encoding="ascii") -> str:
        """Converts native to string

        Args:
            string (Union[str, bytes]): The string to convert
            encoding (str, optional): The encoding type. Defaults to "ascii".

        Returns:
            str: The converted string
        """

        if isinstance(string, str):
            return string

        return string.decode(encoding)

    def get_basic(self, /) -> str:
        """Gets the basic auth header value

        Returns:
            str: The basic auth header value
        """

        username = self.username.encode("latin1")
        password = self.password.encode("latin1")

        return self.to_native_string(b64encode(b":".join((username, password))).strip())



"""Houses Query Option Collection"""

from urllib.parse import urlencode

from pyrestsdk.type.model._options_collection import OptionsCollection
from pyrestsdk.type.model._query_option import QueryOption

class QueryOptionCollection(OptionsCollection[QueryOption]):
    """Query Option Collection Type"""

    def __str__(self) -> str:

        return urlencode(self.as_dict)



"""Houses access token"""

from pyrestsdk.type.model.token._base_token import BaseToken


class AccessToken(BaseToken):
    """Access Token Type"""

    _expires_on: int

    @property
    def expires_on(self) -> int:
        """The token's expiration time in Unix time.
        """
        return self._expires_on



from typing import TypeVar

from pyrestsdk.type.model import Entity
from pyrestsdk.type.exception import UnexpectedReturnType
from pyrestsdk.request.supports_types._supports_base_invoke_request import SupportsBaseInvokeRequest

T = TypeVar("T", bound=Entity)
S = TypeVar("S", bound="SupportsInvokeRequest")

class SupportsInvokeRequest(SupportsBaseInvokeRequest):
    """Supports Invoke Request
    
    Request supports invokation at later time
    """

    @property
    def invoke_request(self:S) -> T:
        """Invokes Request Expecting single value return type

        Args:
            self (S): Supports Invoke Request Instance

        Raises:
            UnexpectedReturnType: Error when return type is other than generic

        Returns:
            T: The single value return
        """

        _return = self.Send(self.input_object)

        if not isinstance(_return, self.generic_type) or _return is None:
            raise UnexpectedReturnType(type(_return), type(self.generic_type))

        return _return



"""Houses Base Request"""

from __future__ import annotations

from typing import (
    TypeVar,
    Union,
    Optional,
    Iterable,
    Dict,
    Any,
)

import logging

from requests import Response

from pyrestsdk.type.enum import HttpsMethod
from pyrestsdk.type.model import (
    Entity,
    QueryOption,
    HeaderOption,
)

from pyrestsdk.request._request import Request

Logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Entity)
Q = TypeVar("Q", bound=QueryOption)
H = TypeVar("H", bound=HeaderOption)


class BaseRequest(Request[T]):
    """The Base Request Type"""

    def _parse_options(self, options: Optional[Iterable[Union[Q, H]]]) -> None:
        """Parses the provided options into either header or query options"""

        Logger.info("%s._parse_options: function called", type(self).__name__)

        if options is None:
            return None

        for option in options:
            match type(option):
                case n if issubclass(n, HeaderOption):
                    self.header_options.append(option)
                case n if issubclass(n, QueryOption):
                    self.query_options.append(option)
                case other:
                    raise TypeError(
                        "Unexpected type: %s, expected subtype of HeaderOption or QueryOption",
                        type(other),
                    )

    def _send_request(
        self, args: Dict[str, Any], value: Optional[Union[T, Dict[str, Any], str]]
    ) -> Optional[Response]:
        """Sends request

        Args:
            args (Dict[str, Any]): _description_
            value (Optional[Union[T, Dict[str, Any], str]]): _description_

        Raises:
            Exception: _description_

        Returns:
            Optional[Response]: Response to request
        """

        Logger.info(
            "%s._send_request: %s request made",
            type(self).__name__,
            self.request_method.name,
        )

        match self.request_method:
            case HttpsMethod.GET:
                return self._client.get(**args)
            case HttpsMethod.POST:
                return self._client.post(**args)
            case HttpsMethod.DELETE:
                self._client.delete(**args)
                return None
            case HttpsMethod.PUT:
                return self._client.put(**args)
            case other:
                raise TypeError(f"Unknown HTTPS method {other.name}")



"""Houses Request Type
"""

from typing import (
    TypeVar,
    List,
    Union,
    Optional,
    Iterable,
    Dict,
    Any,
)

import logging

from urllib.parse import urlparse

import json

from pyrestsdk import AbstractServiceClient

from pyrestsdk.type.enum import HttpsMethod
from pyrestsdk.type.model import (
    Entity,
    QueryOption,
    HeaderOption,
)
from pyrestsdk.request._abstract_request import AbstractRequest
from pyrestsdk.request.supports_types import (
    SupportsGenericType,
    SupportsQueryOptions,
    SupportsHeaderOptions,
)

T = TypeVar("T", bound=Entity)
B = TypeVar("B", bound="Request")
O = TypeVar("O", QueryOption, HeaderOption)
S = TypeVar("S", bound=AbstractServiceClient)

Logger = logging.getLogger(__name__)


class Request(
    SupportsHeaderOptions,
    SupportsQueryOptions,
    SupportsGenericType,
    AbstractRequest[T],
):
    """Request Type
    """

    def __init__(
        self: B, request_url: str, client: S, options: Optional[Iterable[O]]
    ) -> None:
        super().__init__(request_url, client, options)

        self._request_url: str = request_url
        self._client: S = client
        self._method: HttpsMethod = HttpsMethod.GET
        self._request_url: str = self._initialize_url(request_url)
        self._parse_options(options)

    @property
    def request_method(self) -> HttpsMethod:
        """Gets/Sets the https method"""
        return self._method

    @request_method.setter
    def request_method(self, value: HttpsMethod) -> None:
        self._method = value
        Logger.info("%s.Method: request method set to %s", type(self).__name__, value)

    @property
    def request_url(self: B) -> str:
        """Gets/Sets the request URL

        Returns:
            str: The request URL
        """

        return self._request_url

    @request_url.setter
    def request_url(self: B, value: str) -> None:
        self._request_url = value
        Logger.info("%s.request_url: request URL set to %s", type(self).__name__, value)

    @property
    def client(self: B) -> S:
        """Gets the Client"""

        return self._client

    def _initialize_url(self, request_url: str) -> str:
        """Parses the query parameters from URL"""

        Logger.info("%s._initialize_url: function called", type(self).__name__)

        if not request_url:
            raise TypeError("request url can't be None")

        url = urlparse(request_url)

        if url.query:
            self._split_query_options(url.query)

        return url._replace(query="").geturl()

    def send(
        self, __object: Optional[Union[T, Dict[str, Any]]]
    ) -> Optional[Union[List[T], T]]:
        """Submits the request and returns the expected return"""

        Logger.info("%s.Send: method called", type(self).__name__)

        return self.send_request(__object)

    def send_request(
        self, value: Optional[Union[T, Dict[str, Any], str]] = None
    ) -> Optional[Union[List[T], T]]:
        """Makes the desired request and returns the desired return type"""

        Logger.info("%s.SendRequest: method called", type(self).__name__)

        args = self._get_request_args(value)

        _response = self._send_request(args, value)

        if _response is None:
            return None

        try:
            Logger.debug(
                "%s.SendRequest: raising response for status", type(self).__name__
            )
            _response.raise_for_status()
        except Exception:
            self.parse_exception(_response)
        else:
            return self.parse_response(_response)

    def _parse_input_object(self, value: Union[T, Dict[str, Any], str]) -> str:
        """Converts input object into JSON

        Args:
            value (Union[T, Dict[str, Any], str]): The input object

        Returns:
            str: JSON version of object
        """

        if isinstance(value, str):
            return value

        if not isinstance(value, dict):
            value = value.as_dict

        return json.dumps(value)

    def _get_request_args(
        self, value: Optional[Union[T, Dict[str, Any], str]] = None
    ) -> Dict[str, Any]:

        args = {
            "url": self.request_url,
            "headers": self.header_options.as_dict,
            "params": str(self.query_options),
        }

        if (self.request_method == HttpsMethod.POST) or (
            self.request_method == HttpsMethod.PUT
        ):
            if value is None:
                raise TypeError(f"Missing data for {self.request_method.name} request.")

            args["data"] = self._parse_input_object(value)

        return args

    def append_segment_to_request_url(self, url_segment: str) -> str:
        """Gets a URL that is the request builder's request URL with the segment appended.

        Args:
            url_segment (str): The segment to append to the request URL.

        Returns:
            str: A URL that is the request builder's request URL with the segment appended.
        """

        if not url_segment.startswith("/"):
            url_segment = f"/{url_segment}"

        return f"{self.request_url}{url_segment}"



"""Houses enum types"""

from pyrestsdk.type.enum._feature_usage_flag import FeatureUsageFlag
from pyrestsdk.type.enum._https_method import HttpsMethod
from pyrestsdk.type.enum._media_type import MimeType

__all__ = [
    "FeatureUsageFlag",
    "HttpsMethod",
    "MimeType",
]
